#if !defined HAVE_ARRAY_H__
#define      HAVE_ARRAY_H__

#include "fxttypes.h"
#include "bits2pow.h"  // ldq()
#include "fxttypes.h"
#include "newop.h"

#include "inline.h"
#include "copy.h"
#include "shift.h"
#include "minmax.h"
//#include "scan.h"
//#include "scanbox.h"
//#include "misc.h"



template <typename Type>
class array
{
public:
    array(ulong _n, Type *_f=0);
    ~array();

    Type *operator [] (ulong x) const  { return f_[x]; }

    ulong nelem()  const  { return n; }

    void null()  { ::null(f_,n); }
    void copy(array<Type> a)  { ::copy(a.f_, a.n, f_,n); }
    void min_max(Type *mi, Type *ma)  const  { ::min_max(f_,n, mi,ma); }

    void linear_scale(Type mi, Type ma, Type nmi, Type nma)
    { ::linear_scale(f_,n, mi,ma,nmi,nma); }
    void linear_scale(Type nmi, Type nma)
    { ::linear_scale(f_,n, nmi,nma); }
    void log_scale(double xi, double xa, Type nma)
    { ::log_scale(f_,n, xi,xa,nma); }

    void apply_func(Type (*func)(Type)) { ::apply_func(f_,n, func); }

    void shift_left(ulong r)  { ::shift_left(f_,n, r); }
    void shift_right(ulong r)  { ::shift_right(f_,n, r); }
    void shift(long vert, long horiz)  { ::shift(f_,n, vert,horiz); }

    void reverse()  { ::reverse(f_,n); }

    void rotate_left(ulong r)  { ::rotate_left(f_,n, r); }
    void rotate_right(ulong r)  { ::rotate_right(f_,n, r); }
    void rotate(long r)  { ::rotate(f_,n, r); }

//    void zero2center()  { ::zero2center(f_,n); }

    // --- data:
protected:
    const ulong n;
    const ulong ldn;
    bool myfq;

public:
    Type *f_;
};
//---------------------------------------------



template <typename Type>
array<Type>::array(ulong _n, Type *_f/*=0*/)
    : n(_n), ldn(ldq(_n))
{
    if ( 0!=_f )
    {
        f_ = _f;
        myfq = false;
    }
    else
    {
        f_ = NEWOP(Type, n);
        myfq = true;
    }
}
//-----------------------


template <typename Type>
array<Type>::~array()
{
    if ( myfq )  delete [] f_;
}
//-----------------------



template <typename Type>
void
copy(const array<Type> &src,
     array<Type> &dst,
     ulong off/*=0*/)
{
    dst.null();
    ::offset_copy(src.f_, src.n,
                  dst.f_, dst.n,
                  off);
}
//-----------------------



#endif  // !defined HAVE_ARRAY_H__
