#if !defined HAVE_ARRAY2D_H__
#define      HAVE_ARRAY2D_H__

#include "fxttypes.h"
#include "bits2pow.h"  // ldq()

//#include "auxtempl2d.h"
#include "copy2d.h"
#include "shift2d.h"
#include "reverse2d.h"
#include "rotate2d.h"
#include "transpose.h"
#include "scale2d.h"
#include "minmax2d.h"
#include "misc2d.h"
#include "newop.h"


// whether to make range checks on index of requested row
#define  ROW_ASSERT  0  // 0 (fast) or 1 (to catch acces beyond size)

#if  ( ROW_ASSERTS==1 )
// array2d.cc:
void row_assert(ulong r, ulong nr);
#endif

template <typename Type>
class array2d
{
public:
    array2d(ulong _nr, ulong _nc, Type *_f=0);
    array2d(ulong _nr, ulong _nc, Type **_rowp);
    ~array2d();

    ulong nelem()  const  { return n; }
    ulong ncols()  const  { return nc; }
    ulong nrows()  const  { return nr; }

#if  ( ROW_ASSERTS==1 )
    Type *row(ulong r) const  { assert_row(r,nr);  return rowp_[r]; }
    Type *operator [] (ulong y) const  { assert_row(r,nr); return rowp_[y]; }
#else
    Type *row(ulong r) const  { return rowp_[r]; }
    Type *operator [] (ulong y) const  { return rowp_[y]; }
#endif

//    Type **get_rowp()  const  { return rowp_; }

    bool contiguous_mem_q()  const  { return (f_!=0); }


    void null()  { ::null(rowp_,nr,nc); }
    void min_max(Type *mi, Type *ma)  const  { ::min_max((const Type **)rowp_,nr,nc, mi,ma); }
    void linear_scale(Type mi, Type ma, Type nmi, Type nma)  { ::linear_scale(rowp_,nr,nc, mi,ma,nmi,nma); }
    void linear_scale(Type nmi, Type nma)  { ::linear_scale(rowp_,nr,nc, nmi,nma); }
    void log_scale(double xi, double xa, Type nma)  { ::log_scale(rowp_,nr,nc, xi,xa,nma); }

    void apply_func(Type (*func)(Type)) { ::apply_func(rowp_,nr,nc, func); }

    void shift_up(ulong r)  { ::shift_up(rowp_,nr,nc, r); }
    void shift_down(ulong r)  { ::shift_down(rowp_,nr,nc, r); }
    void shift_vert(long r)  { ::shift_vert(rowp_,nr,nc, r); }
    void shift_left(ulong r)  { ::shift_left(rowp_,nr,nc, r); }
    void shift_right(ulong r)  { ::shift_right(rowp_,nr,nc, r); }
    void shift_horiz(long r)  { ::shift_horiz(rowp_,nr,nc, r); }
    void shift(long vert, long horiz)  { ::shift(rowp_,nr,nc, vert,horiz); }

    void reverse_vert()  { ::reverse_vert(rowp_,nr,nc); }
    void reverse_horiz()  { ::reverse_horiz(rowp_,nr,nc); }

    void rotate_up(ulong r)  { ::rotate_up(rowp_,nr,nc, r); }
    void rotate_down(ulong r)  { ::rotate_down(rowp_,nr,nc, r); }
    void rotate_vert(long r)  { ::rotate_vert(rowp_,nr,nc, r); }
    void rotate_left(ulong r)  { ::rotate_left(rowp_,nr,nc, r); }
    void rotate_right(ulong r)  { ::rotate_right(rowp_,nr,nc, r); }
    void rotate_horiz(long r)  { ::rotate_horiz(rowp_,nr,nc, r); }
    void rotate(long vert, long horiz)  { ::rotate(rowp_,nr,nc, vert,horiz); }
    void zero2center()  { ::zero2center(rowp_,nr,nc); }

    void transpose();


    // --- data:
protected:
    const ulong nc, nr, n;
//    const ulong ldnc, ldnr;
    bool myrowpq;  // whether rowp_ was allocated by constructor
    bool myfq;     // whether f_ was allocated by constructor

public:
    Type **rowp_;   // (pointer to) array of pointers to rows
    Type *f_;       // pointer to data if contiguous rows, else 0

};
//---------------------------------------------


template <typename Type>
array2d<Type>::array2d(ulong _nr, ulong _nc, Type *_f=0)
    : nc(_nc), nr(_nr), n(_nc*_nr)
//    , ldnc(ldq(_nc)), ldnr(ldq(_nr))
{
    if ( 0!=_f )
    {
        f_ = _f;
        myfq = false;
    }
    else
    {
        f_ = NEWOP(Type, n);
        myfq = true;
    }

    rowp_ = NEWOP(Type*, nr);
    myrowpq = true;

    rowp_[0] = f_;
    for (ulong i=1; i<nr; ++i)  rowp_[i] = rowp_[i-1] + nc;
}
//-----------------------

template <typename Type>
array2d<Type>::array2d(ulong _nr, ulong _nc, Type **_rowp)
    : nc(_nc), nr(_nr), n(_nc*_nr)
//    , ldnc(ldq(_nc)), ldnr(ldq(_nr))
{
    rowp_ = _rowp;
    myrowpq = false;

    // test if supplied memory is contiguous:
    ulong r = 0;
    Type *p = rowp_[0];
    while ( r<nr )
    {
        ++r;
        p += nc;
        if ( p!=rowp_[r] )  break;
    }

    if ( r==nr )  f_ = rowp_[0];  // memory is contiguous
    else          f_ = 0;

    myfq = false;
}
//-----------------------

template <typename Type>
array2d<Type>::~array2d()
{
    if ( myrowpq )  delete [] rowp_;

    if ( myfq )  delete [] f_;
}
//-----------------------


#if  ( ROW_ASSERTS==1 )
template <typename Type>
Type *
array2d<Type>::row(ulong r) const
{
    if ( r>=nr || (long)r<0 )
    {
        cerr << __PRETTY_FUNCTION__ << endl;
        cerr << " r=" << r << "  (nr=" << nr << ")" << endl;
        assert( r<nr );
        assert( (long)r>=0 );
    }
    return rowp_[r];
}
//-----------------------
#else
// inline member function in array2d.h
#endif


template <typename Type1, typename Type2>
void
copy(const array2d<Type1> &src, array2d<Type2> &dst,
     ulong coff=0, ulong roff=0)
{
    ::offset_copy((const Type1 **)src.rowp_, src.nrows(), src.ncols(),
                  dst.rowp_, dst.nrows(), dst.ncols(),
                  coff, roff);
}
//-----------------------


template <typename Type>
void
array2d<Type>::transpose()
{
    if ( nr==nc )  // ok for square array
    {
        ::transpose(rowp_, nr, nc);
    }
    else  // rectangular array ...  jjtodo: complete transpose for array2d
    {
        jjassert2(0, "sorry, transpose for nonsquare array2d not yet implemented.");
        /*
        if ( contiguous_mem_q() )
        {
            ; // problem: need to realloc rowp_
            // ... or alloc rowp_[ max(nr, nc) ] in constructor
        }
        else
        {
            ; // problem: need to implement transpose for noncontiguous memory
        }
        */
    }
}
//-----------------------

#endif  // !defined HAVE_ARRAY2D_H__
