#if !defined HAVE_FUNCEMU_H__
#define      HAVE_FUNCEMU_H__


#include "fxttypes.h"

template <typename Type>
class funcemu
{
public:
    funcemu(ulong nstates, ulong ndata)
        : ststptr_(0), stptr_(0)
    {
        ststack_ = new ulong[nstates];
        stack_   = new Type[ndata];
    }

    ~funcemu()
    {
        delete [] stack_;
        delete [] ststack_;
    }

    funcemu & operator = (const funcemu &);  // forbidden

    void stpush(ulong x)  { ststack_[ststptr_++] = x; }
    void stpull(ulong &x)  { x = ststack_[--ststptr_]; }
    void stpeek(ulong &x)  { x = ststack_[ststptr_-1]; }
    void stpoke(ulong x)  { ststack_[ststptr_-1] = x; }
    void stpop(ulong ct=1)  { ststptr_-=ct; }


    void push(Type x)  { stack_[stptr_++] = x; }
    void push(Type x, Type y)  { push(x); push(y); }
    void push(Type x, Type y, Type z)  { push(x); push(y); push(z); }
    void push(Type x, Type y, Type z, Type u)
    { push(x); push(y); push(z); push(u); }

    void pull(Type &x)  { x = stack_[--stptr_]; }
    void pull(Type &x, Type &y)  { pull(y); pull(x); }
    void pull(Type &x, Type &y, Type &z)  { pull(z); pull(y); pull(x); }
    void pull(Type &x, Type &y, Type &z, Type &u)
    { pull(u); pull(z); pull(y); pull(x); }

    void peek(Type &x)  { x = stack_[stptr_-1]; }
    void peek(Type &x, Type &y)
    { y = stack_[stptr_-1]; x = stack_[stptr_-2]; }
    void peek(Type &x, Type &y, Type &z)
    { z = stack_[stptr_-1]; y = stack_[stptr_-2]; x = stack_[stptr_-3]; }
    void peek(Type &x, Type &y, Type &z, Type &u)
    {
        u = stack_[stptr_-1];
        z = stack_[stptr_-2];
        y = stack_[stptr_-3];
        x = stack_[stptr_-4];
    }

    void poke(Type x)  { stack_[stptr_-1] = x; }
    void poke(Type x, Type y)
    { stack_[stptr_-1] = y; stack_[stptr_-2] = x; }
    void poke(Type x, Type y, Type z)
    { stack_[stptr_-1] = z; stack_[stptr_-2] = y; stack_[stptr_-3] = x; }
    void poke(Type x, Type y, Type z, Type u)
    {
        stack_[stptr_-1] = u;
        stack_[stptr_-2] = z;
        stack_[stptr_-3] = y;
        stack_[stptr_-4] = x;
    }

    void pop(ulong ct=1) { stptr_-=ct; }


    void reset()  { stptr_=0; ststptr_=0; }

public:
    ulong ststptr_;  // state stack pointer
    ulong stptr_;   // data stack pointer

    ulong *ststack_; // state stack
    Type  *stack_; // data stack
};


#endif // !defined HAVE_FUNCEMU_H__
