#if !defined HAVE_MISC_H__
#define      HAVE_MISC_H__


#include "fxttypes.h"


//#define  UNROLLED_SUMMATION

template <typename Type>
inline Type
get_sum(const Type *f, ulong n)
// return sum of array f[]
{
    Type s = 0;
#if !defined  UNROLLED_SUMMATION

    while ( n-- )  s += f[n];

#else // UNROLLED_SUMMATION

    if ( n & 1 )  s += f[--n];
    if ( n & 2 )  s += f[--n] + f[--n];

    if ( n >= 4 )
    {
        Type s0 = 0, s1 = 0, s2 = 0, s3 = 0; 
        while ( n )
        {
            n  -= 4;
            s0 += f[n];
            s1 += f[n+1];
            s2 += f[n+2];
            s3 += f[n+3];
        }
        s = s0 + s1 + s2 + s3;
    }
#endif // UNROLLED_SUMMATION
    return s;
}
//-----------------------


template <typename Type>
void
diff(Type *f, ulong n, ulong j=1)
// first difference (applied j times) of f[]
{
    if ( j>=n )
    {
//        null(f,n);
        const Type v(0);
        while ( n-- )  f[n] = v;

        return;
    }

    for (ulong k=1; k<=j; ++k)
    {
        for (ulong i=1; i<=n-k; ++i)
        {
            f[i-1] = f[i] - f[i-1];
        }
        f[n-k] = 0;
    }
}
//-----------------------


template <typename Type>
int
compare(const Type *f, const Type *g, ulong n)
// compare f[], g[] up to length n
// if for f[k]!=g[k] for any k<n
// then return sign(f[k]-g[k])
{
    for (ulong k=0; k<n; ++k)
    {
        if ( f[k]==g[k] )  continue;

        if ( f[k]>g[k] )  return   1;
        else              return  -1;
    }
    return  0;
}
//-----------------------


template <typename Type>
inline void
negate(Type *f, ulong n)
// negate f[]
{
    while ( n-- )  f[n] = -f[n];
}
//-----------------------


// multiply array by s:
template <typename Type>
inline void
multiply(Type *f, ulong n, Type s)
// multiply f[] by s
{
    while ( n-- )  f[n] *= s;
}
//-----------------------


template <typename Type>
inline void
apply_func(Type *f, ulong n, Type (*func)(Type))
// apply function func to each element of f[]
{
    while ( n-- )  f[n] = func(f[n]);
}
//-----------------------


template <typename Type>
inline ulong
sequency(Type *f, ulong n)
// return the number of sign changes in f[]
{
    ulong sq = 0;
    ulong s1 = (f[0] < 0 ? 0 : 1);
    for (ulong k=1; k<n; ++k)
    {
        ulong s2 = (f[k] < 0 ? 0 : 1);
        sq += (s1 ^ s2);
        s1 = s2;
    }

    return  sq;
}
//-----------------------


#endif // !defined HAVE_MISC_H__
